////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file system_internal.h
/// Declarations of internal system functions
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include "system.h"

#include <allegro5/allegro.h>


////////////////////////////////////////////////////////////////////////////////
// randomizer
////////////////////////////////////////////////////////////////////////////////

#include <gsl/gsl_rng.h>

#define TS5_RNG							gsl_rng


////////////////////////////////////////////////////////////////////////////////
// display
////////////////////////////////////////////////////////////////////////////////

typedef struct {

	int x1;
	int x2;
	int y1;
	int y2;
	int w;
	int h;

} TS5_DISPLAY_ADAPTER_STATUS;


typedef struct {

	int adapter;
	int w;
	int h;
	int x;
	int y;
	int refreshrate;
	double last_sync;
	int display_mode;
	int vsync_mode;
	
} TS5_DISPLAY_STATUS;

extern TS5_DISPLAY_STATUS ts5_nextdisplay;


////////////////////////////////////////////////////////////////////////////////
// graphics
////////////////////////////////////////////////////////////////////////////////

/// Structure that contains the graphics settings.
typedef struct {

	int coordinate_system;
	int coordinate_scale;

	int target_width;
	int target_height;

	TS5_COLOR foreground_color;
	TS5_COLOR background_color;
	double drawing_thickness;
	int fill_mode;

	int font_type;
	int font_style;
	int font_size;
	int font_index;

	int text_alignment;

	TS5_FONT *fonts[TS5_NFONTS];
	TS5_FONT *font;

} TS5_GRAPHICS_STATUS;


////////////////////////////////////////////////////////////////////////////////
// audio
////////////////////////////////////////////////////////////////////////////////

#include <allegro5/allegro_audio.h>

typedef struct {
	unsigned int samplerate;
	unsigned int channels;
	int depth;
	double gain;
} TS5_AUDIO_STATUS;

typedef struct{
	ALLEGRO_VOICE *voice;
	ALLEGRO_MIXER *mixer;	
	ALLEGRO_AUDIO_RECORDER *recorder;
	ALLEGRO_EVENT_QUEUE *recorder_queue;
} TS5_AUDIO_DATA;

typedef struct {
	int num_buttons;
	int num_defined_buttons;
	int num_active_buttons;
	double treshold;
	double rise_delay;
	double drop_delay;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
} TS5_VOICEKEY_STATUS;

////////////////////////////////////////////////////////////////////////////////
// mouse
////////////////////////////////////////////////////////////////////////////////

typedef struct {
	int num_buttons;
	int num_defined_buttons;
	int num_active_buttons;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
} TS5_MOUSE_STATUS;


////////////////////////////////////////////////////////////////////////////////
// keyboard
////////////////////////////////////////////////////////////////////////////////

typedef struct {
	int num_buttons;
	int num_defined_buttons;
	int num_active_buttons;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
} TS5_KEYBOARD_STATUS;


////////////////////////////////////////////////////////////////////////////////
// joystick
////////////////////////////////////////////////////////////////////////////////

typedef struct {
	int num_buttons;
	int num_defined_buttons;
	int num_active_buttons;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
} TS5_JOYSTICK_STATUS;


////////////////////////////////////////////////////////////////////////////////
// cedrusbox
////////////////////////////////////////////////////////////////////////////////

typedef struct {
	int type;
	int port_num;
	int num_buttons;
	int num_defined_buttons;
	int num_active_buttons;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
	
	int is_trigger_input_device;
	int simulate_trigger_input;
	double trigger_simulation_interval;
	double last_trigger_simulation;
	
} TS5_CEDRUSBOX_STATUS;


////////////////////////////////////////////////////////////////////////////////
// parport
////////////////////////////////////////////////////////////////////////////////

typedef struct {
	int num_buttons;
	int num_defined_buttons;
	int num_active_buttons;
	double button_debounce_time;
	int *status_register_current_state;
	int *status_register_previous_state;
	double *status_register_last_state_toggle_time;
	int *button_press_defined;
	int *button_press_active;
	int *button_release_defined;
	int *button_release_active;
	int is_trigger_input_device;
	int is_trigger_output_device;
	double trigger_debounce_time;
	int simulate_trigger_input;
	double trigger_simulation_interval;
	double last_trigger_simulation;
	int data_register_current_state;
	int data_register_previous_state;
	double data_register_last_state_toggle_time;
	double trigger_output_time;
} TS5_PARPORT_STATUS;


////////////////////////////////////////////////////////////////////////////////
// serialport
////////////////////////////////////////////////////////////////////////////////

typedef struct {
	char portname[TS5_MAX_CHAR];
	TS5_SERIALPORT *port;
	int is_trigger_input_device;
	int is_trigger_output_device;
	int simulate_trigger_input;
	double trigger_simulation_interval;
	double last_trigger_simulation;
} TS5_SERIALPORT_STATUS;


////////////////////////////////////////////////////////////////////////////////
// timer
////////////////////////////////////////////////////////////////////////////////

typedef struct {

	int priority;
	
	int num_defined_buttons;
	int num_active_buttons;

	int voicekey_is_response_device;
	TS5_VOICEKEY_STATUS voicekey;

	int mouse_is_response_device;
	TS5_MOUSE_STATUS mouse;

	int keyboard_is_response_device;
	TS5_KEYBOARD_STATUS keyboard;

	int num_joysticks;
	int joystick_is_response_device;
	TS5_JOYSTICK_STATUS *joystick;

	int num_cedrusboxes;
	int cedrusbox_is_response_device;
	int cedrusbox_is_trigger_device;
	TS5_CEDRUSBOX_STATUS *cedrusbox;

	int num_parports;
	short parport_base_address[TS5_PARPORT_MAXPORT];
	short parport_data_register[TS5_PARPORT_MAXPORT];
	short parport_status_register[TS5_PARPORT_MAXPORT];
	short parport_control_register[TS5_PARPORT_MAXPORT];
	int parport_is_response_device;
	int parport_is_trigger_device;
	TS5_PARPORT_STATUS *parport;

	int num_serialports;
	int serialport_is_trigger_device;
	TS5_SERIALPORT_STATUS *serialport;

} TS5_TIMER_STATUS;

typedef struct {

	ALLEGRO_EVENT_SOURCE voicekey_response_event_source;

	ALLEGRO_EVENT_SOURCE cedrusbox_response_event_source;
	ALLEGRO_EVENT_SOURCE cedrusbox_trigger_event_source;
	ALLEGRO_THREAD *cedrusbox_thread;

	ALLEGRO_EVENT_SOURCE parport_response_event_source;
	ALLEGRO_EVENT_SOURCE parport_trigger_event_source;
	ALLEGRO_THREAD *parport_thread;

	ALLEGRO_EVENT_SOURCE serialport_trigger_event_source;
	ALLEGRO_THREAD *serialport_thread;

	ALLEGRO_EVENT_QUEUE *response_queue;
	ALLEGRO_EVENT_QUEUE *trigger_queue;
	ALLEGRO_EVENT_QUEUE *trigger_log;
	
	int realtime_clock;

} TS5_TIMER_DATA;


////////////////////////////////////////////////////////////////////////////////
// system
////////////////////////////////////////////////////////////////////////////////

#define TS5_TEXTLOG						ALLEGRO_TEXTLOG

extern int  _ts5_is_tscope5_installed;

extern int  _ts5_is_textlog_installed;


/// Structure that contains the complete status of tscope.
typedef struct {

	unsigned int loglevel;
	int num_display_adapters;
	TS5_DISPLAY_ADAPTER_STATUS *display_adapter;
	int num_displays;
	int active_display;
	TS5_DISPLAY_STATUS *display;
	TS5_GRAPHICS_STATUS graphics;
	TS5_AUDIO_STATUS audio;
	TS5_TIMER_STATUS timer;

} TS5_STATUS;

TS5_STATUS _ts5_status;


typedef struct {
	
	FILE *textlog;

	TS5_RNG *randomizer;

	TS5_DISPLAY **display;
	TS5_BITMAP **display_buffer;
	TS5_BITMAP **display_backup;
	TS5_BITMAP *target;
		
	TS5_AUDIO_DATA audio;
	TS5_TIMER_DATA timer;
	
} TS5_DATA;

TS5_DATA _ts5_data;


#ifdef __cplusplus
extern "C" {
#endif
	
void ts5_install_tscope5(char *calling_function);
void ts5_uninstall_tscope5();

#ifdef __cplusplus
}
#endif

