////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file bitmaps_internal.c
/// Definitions of internal bitmap functions.
////////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/bitmaps_internal.h"
#include "../include/tscope5/system_internal.h"

#include <allegro5/allegro_image.h>

/// Is the bitmaps subsystem installed?
int _ts5_is_bitmaps_installed = 0;


////////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each bitmap function.
///
/// \param calling_function Name the function that calls for 
/// this check or installation.
///
/// Checks whether the bitmaps subsystem is installed. 
/// If not, the bitmaps subsystem is installed.
////////////////////////////////////////////////////////////////////////////////
void ts5_check_bitmaps(char *calling_function)
{
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_bitmaps\n", calling_function);
	if (!_ts5_is_bitmaps_installed) {
		ts5_install_bitmaps(calling_function);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each bitmap function.
///
/// \param calling_function Name the function that calls for 
/// this check or installation.
///
/// \param map The bitmap that has to be checked.
///
/// Checks whether the source and target bitmap are not NULL or equal.
/// Aborts if necessary.
////////////////////////////////////////////////////////////////////////////////
void ts5_check_bitmaps2(char *calling_function, TS5_BITMAP *map)
{
	ts5_check_bitmaps(calling_function);
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_bitmaps2\n", calling_function);

	if (!_ts5_data.target) {
		ts5_fatal("%s: no drawing target specified\n", calling_function);
	}

	if (!map) {
		ts5_fatal("%s: bitmap pointer is null\n", calling_function);
	}

	if (_ts5_data.target == map) {
		ts5_fatal("%s: source and target are the same\n", calling_function);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Install the bitmaps subsystem.
///
/// \param calling_function Name the function that calls for 
/// this check or installation.
///
/// This function is called automatically if necessary.
////////////////////////////////////////////////////////////////////////////////
void ts5_install_bitmaps(char *calling_function)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5(calling_function);
	}

	if (!_ts5_is_bitmaps_installed) {

		_ts5_is_bitmaps_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 bitmaps\n", 
				calling_function);

		if (!al_init_image_addon()) {
			ts5_fatal("%s: could not install Tscope5 bitmaps\n", 
				calling_function);
		}

		atexit(ts5_uninstall_bitmaps);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Close the bitmaps subsystem.
///
/// This function is called automatically at the end of the program.
////////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_bitmaps()
{
	if (_ts5_is_bitmaps_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 bitmaps\n");
		al_shutdown_image_addon();
		_ts5_is_bitmaps_installed = 0;
	}
}

