////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file display.c
/// Definitions of display functions.
/// \example display01.c
/// \example display01.cpp
/// \example display02.c
/// \example display03.c
/// \example display04.c
/// \example display05.c
////////////////////////////////////////////////////////////////////////////////


#include "../include/tscope5/display.h"
#include "../include/tscope5/graphics.h"
#include "../include/tscope5/bitmaps.h"
#include "../include/tscope5/display_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/graphics_internal.h"

#include <math.h>


////////////////////////////////////////////////////////////////////////////////
/// @name Basic operations
/// A display can either be a window or a fullscreen display.
/// There are three basic display operations: 
/// - You will need to install a display with ts5_install_display() 
///   before anything can be drawn.
/// - The contents of a display can be cleared with ts5_clear_display().
/// - Anything that is drawn on a display only becomes visible 
///   after a call to ts5_flip_display().
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Open a new display.
///
/// \return The index of the new display.
///
/// TS5_FULLSCREEN_WINDOW mode:
/// - A fullscreen display will be opened using the same 
///  display size and refreshrate as the desktop.
/// - In this mode, the display size, position and refreshrate 
///   cannot be set by the user. These parameters will be ignored.
/// - Tscope5 will check whether the new display is the first 
///   display on the adapter. If not, the program is aborted.
///
/// TS5_FULLSCREEN mode:
/// - A fullscreen display will be opened. 
///   The user can specify the display size and refreshrate.
/// - In this mode, the display position cannot be set by the user. 
///   This parameter will be ignored.
/// - Tscope5 will check whether the new display is the 
///   first display on the adapter. If not, the program is aborted.
/// - Tscope5 will check whether the display size and refreshrate is 
///   supported by the adapter. If not, the program is aborted.
/// - A call to ts5_print_status() will list all combinations of 
///  size and refreshrate that are supported by the adapter.
///
/// TS5_WINDOWED mode:
/// - Opens a window. The user can specifiy the display size and position.
/// - In this mode, the refreshrate cannot be set by the user. 
///   This parameter will be ignored.
/// - Tscope5 will check whether the adapter is already in use by 
///   a TS5_FULLSCREEN_WINDOW or TS5_FULLSCREEN display. 
///   If so, the program is aborted.
/// - Tscope5 will check whether the window will fit on the display. 
///   If not, the dimensions of the display are reduced.
/// - Tscope5 will check whether the window will be visible on the screen. 
///   If not, the window position is changed.
///
/// The new display is the drawing target for future drawing operations.
////////////////////////////////////////////////////////////////////////////////
int ts5_install_display()
{
	ts5_check_display("ts5_install_display");

	TS5_DISPLAY_STATUS request = ts5_nextdisplay;

	// TS5_FULLSCREEN_WINDOW
	if (ts5_nextdisplay.display_mode == TS5_FULLSCREEN_WINDOW) {

		// the adapter cannot be in use by another tscope5 display
		int i;
		for (i = 0; i < _ts5_status.num_displays; i++) {

			if (_ts5_status.display[i].adapter 
					== ts5_nextdisplay.adapter) {
				ts5_fatal ("%s: %s, %s (%d)\n", "ts5_install_display",
							"cannot open fullscreen display",
							"adapter already in use by another display", i);
			}
		}

		// copy relevant settings
		request.w = 320;
		request.h = 240;
		request.x = -1;
		request.y = -1;
		request.refreshrate = -1;
		request.display_mode = TS5_FULLSCREEN_WINDOW;
		request.vsync_mode = ts5_nextdisplay.vsync_mode;
		
		#ifdef TS5_RASPBERRYPI
		
			request.display_mode = TS5_FULLSCREEN;
			request.w = ts5_get_display_adapter_width(1);
			request.h = ts5_get_display_adapter_height(1);
			
						
		#endif
	}


	// TS5_FULLSCREEN
	if (ts5_nextdisplay.display_mode == TS5_FULLSCREEN) {

		// the adapter cannot be in use by another tscope5 display
		int i;
		for (i = 0; i < _ts5_status.num_displays; i++) {

			if (_ts5_status.display[i].adapter 
					== ts5_nextdisplay.adapter) {
				ts5_fatal ("%s: %s, %s (%d)\n", "ts5_install_display",
							"cannot open fullscreen display",
							" adapter already in use by another display", i);
			}
		}

		// is the combination of display size and refresh rate supported by 
		// the adapter?
		ALLEGRO_DISPLAY_MODE *mode;
		mode = (ALLEGRO_DISPLAY_MODE *)al_malloc(sizeof(ALLEGRO_DISPLAY_MODE));
		int nmodes = al_get_num_display_modes();
		int found=0;

		for (i=0; i<nmodes; i++) {

			mode = al_get_display_mode(i, mode);

			if (mode->width==ts5_nextdisplay.w 
					&& mode->height==ts5_nextdisplay.h 
					&& (mode->refresh_rate==ts5_nextdisplay.refreshrate 
					|| mode->refresh_rate==0)) {
				found=1;
			}
		}

		al_free(mode);

		if (!found) {
			ts5_fatal ("%s: %s (%d, %d) %s (%d) %s %d\n",
						"ts5_install_display",
						"the requested combination of size",
						ts5_nextdisplay.w, ts5_nextdisplay.h,
						"and refreshate", ts5_nextdisplay.refreshrate,
						"is not supported by display adapter",
						ts5_nextdisplay.adapter);
		}

		// copy relevant settings
		request.w = ts5_nextdisplay.w;
		request.h = ts5_nextdisplay.h;
		request.x = -1;
		request.y = -1;
		request.refreshrate = ts5_nextdisplay.refreshrate;
		request.display_mode = TS5_FULLSCREEN;
		request.vsync_mode = ts5_nextdisplay.vsync_mode;
	}

	// TS5_WINDOWED
	if (ts5_nextdisplay.display_mode == TS5_WINDOWED) {

		// the adapter cannot be in used by a 
		// TS5_FULLSCREEN_WINDOW or TS5_FULLSCREEN display
		int i;
		for (i = 0; i < _ts5_status.num_displays; i++) {

			if (_ts5_status.display[i].adapter 
					== ts5_nextdisplay.adapter 
					&& (_ts5_status.display[i].display_mode 
					== TS5_FULLSCREEN 
					|| _ts5_status.display[i].display_mode 
					== TS5_FULLSCREEN_WINDOW)) {

				ts5_fatal ("%s: %s, %s (%d)\n", "ts5_install_display",
						 	"cannot open display",
							" adapter already in use by a fullscreen display",
							i);
			}
		}

		// window should be at least 320 pixels wide 
		// and smaller than the width of the monitor
		if (ts5_nextdisplay.w < 320 ) {
			request.w = 320;
			ts5_log(TS5_LOGLEVEL_1, "%s: %s %d (was %d)\n",
						"ts5_install_display", "adjusting display width to",
						request.w, ts5_nextdisplay.w);
		}
		else if (ts5_nextdisplay.w 
					> _ts5_status.display_adapter
					[ts5_nextdisplay.adapter].w) {
							
			request.w = _ts5_status.display_adapter[ts5_nextdisplay.adapter].w;
			
			ts5_log(TS5_LOGLEVEL_1, "%s: %s %d (was %d)\n",
						"ts5_install_display", "adjusting display width to", 
						request.w, ts5_nextdisplay.w);
		}
		else {
			request.w = ts5_nextdisplay.w;
		}

		// window should be at least 240 pixels high 
		// and smaller than the height of the monitor
		if (ts5_nextdisplay.h < 240 ) {
			request.h = 240;
			ts5_log(TS5_LOGLEVEL_1, "%s: %s %d (was %d)\n",
						"ts5_install_display", "adjusting display height to",
						request.h, ts5_nextdisplay.h);
		}
		else if (ts5_nextdisplay.h 
					> _ts5_status.display_adapter
					[ts5_nextdisplay.adapter].h) {
						
			request.h = _ts5_status.display_adapter
						[ts5_nextdisplay.adapter].h;
			
			ts5_log(TS5_LOGLEVEL_1, "%s: %s %d (was %d)\n",
						"ts5_install_display", "adjusting display height to",
						request.h, ts5_nextdisplay.h);
		}
		else {
			request.h = ts5_nextdisplay.h;
		}

		request.x = ts5_nextdisplay.x;
		request.y = ts5_nextdisplay.y;

		if (request.x == -1) {

			if (_ts5_status.graphics.coordinate_system 
					== TS5_CARTESIAN_COORDINATES) {
						
				request.x = (_ts5_status.display_adapter
								[request.adapter].w - request.w)/2;
			}
			else {
				request.x = 0;
			}

			request.x += _ts5_status.display_adapter[request.adapter].x1;
		}

		if (request.y == -1) {

			if (_ts5_status.graphics.coordinate_system 
					== TS5_CARTESIAN_COORDINATES) {
						
				request.y = (_ts5_status.display_adapter
								[request.adapter].h - request.h)/2;
			}
			else {
				request.y = 0;
			}

			request.y += _ts5_status.display_adapter[request.adapter].y1;
		}

		// copy relevant settings
		request.refreshrate = -1;
		request.display_mode = TS5_WINDOWED;
		request.vsync_mode = ts5_nextdisplay.vsync_mode;
	}


	// pass on display parameters to Allegro
	al_set_new_display_adapter(request.adapter);
	al_set_new_display_refresh_rate(request.refreshrate);
	
	al_set_new_display_option(ALLEGRO_VSYNC, (int)request.vsync_mode, 
								ALLEGRO_SUGGEST);

	if (request.display_mode == TS5_WINDOWED) {
		al_set_new_window_position(request.x, request.y);
	}

	if (request.display_mode == TS5_FULLSCREEN_WINDOW) {
		al_set_new_window_position(INT_MAX, INT_MAX);
	}

	al_set_new_display_flags(request.display_mode);

	// update counters
	_ts5_status.num_displays++;
	_ts5_status.active_display = _ts5_status.num_displays - 1;

	// try top open display
	ts5_log(TS5_LOGLEVEL_1, "ts5_install_display: Opening display\n");

	TS5_DISPLAY **display;
	display = al_realloc(_ts5_data.display, 
							_ts5_status.num_displays * sizeof(TS5_DISPLAY *));

	if (display == NULL) {
		ts5_fatal("%s: %s\n", "ts5_install_display",
				 	"failed to allocate memory for display pointer");
	}

	_ts5_data.display = (TS5_DISPLAY **) display;
	
	_ts5_data.display[_ts5_status.active_display] = 
		al_create_display(request.w, request.h);

	if (!_ts5_data.display[_ts5_status.active_display]) {
		ts5_fatal("ts5_install_display: could not open display\n");
	}

	if (ts5_nextdisplay.display_mode == TS5_FULLSCREEN_WINDOW) {
		request.w = _ts5_status.display_adapter[request.adapter].w;
		request.h = _ts5_status.display_adapter[request.adapter].h;
	}

	// make a display buffer
	TS5_BITMAP **display_buffer;

	display_buffer = al_realloc(_ts5_data.display_buffer, 
						_ts5_status.num_displays * sizeof(TS5_BITMAP *));

	if (display_buffer == NULL) {
		ts5_fatal("%s: %s\n", "ts5_install_display",
				 	"failed to allocate memory for display buffer pointer");
	}

	_ts5_data.display_buffer = (TS5_BITMAP **) display_buffer;
	
	_ts5_data.display_buffer[_ts5_status.active_display] = 
		al_create_bitmap(request.w, request.h);

	if (!_ts5_data.display_buffer[_ts5_status.active_display]) {
		ts5_fatal("ts5_install_display: could not open display buffer\n");
	}


	// make a backup of the display buffer
	TS5_BITMAP **display_backup;

	display_backup = al_realloc(_ts5_data.display_backup, 
						_ts5_status.num_displays * sizeof(TS5_BITMAP *));

	if (display_backup == NULL) {
		ts5_fatal("%s: %s\n", "ts5_install_display",
				 	"failed to allocate memory for display backup pointer");
	}

	_ts5_data.display_backup = (TS5_BITMAP **) display_backup;
	
	_ts5_data.display_backup[_ts5_status.active_display] = 
		al_create_bitmap(request.w, request.h);

	if (!_ts5_data.display_backup[_ts5_status.active_display]) {
		ts5_fatal("ts5_install_display: could not open display backup\n");
	}


	// copy the settings in the  structure
	TS5_DISPLAY_STATUS *display_status;
	display_status = al_realloc(_ts5_status.display, 
						_ts5_status.num_displays * sizeof(TS5_DISPLAY_STATUS));

	if (display_status == NULL) {
		ts5_fatal("%s: %s\n", "ts5_install_display",
				 	"failed to allocate memory for display settings");
	}

	_ts5_status.display = (TS5_DISPLAY_STATUS *) display_status;
	
	_ts5_status.display[_ts5_status.active_display].adapter = 
		request.adapter;

	if (request.display_mode==TS5_WINDOWED) {

		al_get_window_position(_ts5_data.display[_ts5_status.active_display], 
								&request.x, &request.y);

		if (request.x + request.w 
			> _ts5_status.display_adapter[ts5_nextdisplay.adapter].x2) {

			int oldw = request.w;
			
			request.w = _ts5_status.display_adapter
							[ts5_nextdisplay.adapter].x2 - request.x;
			
			al_resize_display(_ts5_data.display[_ts5_status.active_display], 
								request.w, request.h);
			
			ts5_log(TS5_LOGLEVEL_1, "%s: %s %d (was %d)\n",
						"ts5_install_display", "adjusting display width to",
						request.w, oldw);
		}

		if (request.y + request.h > _ts5_status.display_adapter
				[ts5_nextdisplay.adapter].y2) {

			int oldh = request.h;
			
			request.h = _ts5_status.display_adapter
						[ts5_nextdisplay.adapter].y2 - request.y;
			
			al_resize_display(_ts5_data.display[_ts5_status.active_display], 
								request.w, request.h);
			
			al_set_window_position(_ts5_data.display
									[_ts5_status.active_display],
									request.x, request.y + (request.h-oldh));

			ts5_log(TS5_LOGLEVEL_1, "%s: %s %d (was %d)\n",
						"ts5_install_display", "adjusting display height to",
						request.h, oldh);

			al_get_window_position(
				_ts5_data.display[_ts5_status.active_display], 
				&request.x, &request.y);
		}

		request.x -= _ts5_status.display_adapter[request.adapter].x1;
		request.y -= _ts5_status.display_adapter[request.adapter].y1;

		if (_ts5_status.graphics.coordinate_system 
				== TS5_CARTESIAN_COORDINATES) {
			
			request.x = 
				request.x 
				- (_ts5_status.display_adapter[request.adapter].w 
				- request.w)/2.0;
			
			request.y = 
				(_ts5_status.display_adapter[request.adapter].h 
				- request.h)/2.0 
				- request.y;
		}
	}

	_ts5_status.display[_ts5_status.active_display].x = request.x;
	_ts5_status.display[_ts5_status.active_display].y = request.y;
	_ts5_status.display[_ts5_status.active_display].w = request.w;
	_ts5_status.display[_ts5_status.active_display].h = request.h;

	_ts5_status.display[_ts5_status.active_display].refreshrate = 
		ts5_nextdisplay.refreshrate;

	_ts5_status.display[_ts5_status.active_display].display_mode = 
		request.display_mode;

	_ts5_status.display[_ts5_status.active_display].vsync_mode = 
		request.vsync_mode;

	// set the new display as the drawing target and clear the display
	ts5_set_drawing_target(_ts5_data.display_backup
							[_ts5_status.active_display]);
	al_clear_to_color(_ts5_status.graphics.background_color);

	ts5_set_drawing_target(_ts5_data.display_buffer
							[_ts5_status.active_display]);
		
	
	// clear and flip
	int i;
	for(i=0; i<4; i++) {
		al_flip_display();	
	}
	_ts5_status.display[_ts5_status.active_display].last_sync = al_get_time();
	ts5_clear_display();
	ts5_flip_display();

	// set atexit and return
	atexit(ts5_uninstall_displays);

	return _ts5_status.active_display;
}


////////////////////////////////////////////////////////////////////////////////
/// Clear the current drawing target.
///
/// The drawing target can either be a display or a bitmap (to clear a bitmap:
/// set it as drawing target and call ts5_clear_display()).
///
/// The background color is controlled by ts5_set_background_color().
////////////////////////////////////////////////////////////////////////////////
void ts5_clear_display()
{
	ts5_check_display("ts5_clear_display");
	ts5_log(TS5_LOGLEVEL_5, "ts5_clear_display()\n");

	if (_ts5_data.target == NULL) {
		ts5_fatal("ts5_clear_display: target is a NULL pointer\n");
	}

	al_clear_to_color(_ts5_status.graphics.background_color);
}


////////////////////////////////////////////////////////////////////////////////
/// Make what has been drawn visible on the screen.
///
/// \return The time when the display flip occurs.
///
/// All drawing operations (primitives, text, bitmaps, ...) 
/// are performed on a back buffer.
/// Once you have finished drawing you should call this function 
/// to make your drawings visible on the screen.
/// After a call to this function the back buffer is not cleared.
////////////////////////////////////////////////////////////////////////////////
double ts5_flip_display()
{
	ts5_check_display("ts5_flip_display");
	ts5_log(TS5_LOGLEVEL_5, "ts5_flip_display()\n");


	// if we do not synchronize with _every_ vsync
	// we need to call al_wait_for_vsync to get in sync again
	#ifdef TS5_WINDOWS
		if (al_get_time() - _ts5_status.display[_ts5_status.active_display].last_sync >
			1.0 / _ts5_status.display[_ts5_status.active_display].refreshrate + 0.003) {
				al_wait_for_vsync();		
		}
	#endif

	// on mac we need another trick for this:
	// blit the backup of the buffer
	// then make a new backup of the buffer
	#ifdef TS5_MACOSX
		if (al_get_time() - _ts5_status.display[_ts5_status.active_display].last_sync >
			1.0 / _ts5_status.display[_ts5_status.active_display].refreshrate + 0.003) {
				
				al_set_target_backbuffer(_ts5_data.display[_ts5_status.active_display]);
				al_draw_bitmap(_ts5_data.display_backup[_ts5_status.active_display], 0.0, 0.0, 0);
				al_flip_display();
		}
		
		al_set_target_bitmap(_ts5_data.display_backup[_ts5_status.active_display]);
		al_clear_to_color(_ts5_status.graphics.background_color);
		al_draw_bitmap(_ts5_data.display_buffer[_ts5_status.active_display], 0.0, 0.0, 0);				
	#endif
	
		
	al_set_target_backbuffer(_ts5_data.display[_ts5_status.active_display]);
	al_draw_bitmap(_ts5_data.display_buffer[_ts5_status.active_display], 0.0, 0.0, 0);
	al_flip_display();
	
	_ts5_status.display[_ts5_status.active_display].last_sync = al_get_time();
	
	al_clear_to_color(_ts5_status.graphics.background_color);
	
	al_set_target_bitmap(_ts5_data.display_buffer
							[_ts5_status.active_display]);
	
	return _ts5_status.display[_ts5_status.active_display].last_sync;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// @name Setting the display target
/// Tscope5 supports multiple displays 
/// (each call to ts5_install_display() opens a new display). 
/// - By default everything is drawn on the display that was opened last.
/// - With ts5_set_active_display() you can change the target display.
///
/// You can also redirect drawing to memory bitmaps (see bitmaps.c) 
/// using ts5_set_drawing_target().
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Set the active display.
///
/// \param display Index of the display.
///
/// \return The index of the previous active display.
///
/// The 'active' display is the display that will be used 
/// for drawing operations.
////////////////////////////////////////////////////////////////////////////////
int ts5_set_active_display(const int display)
{
	ts5_check_display2("ts5_set_active_display", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_active_display(%d)\n", display);

	int retval = _ts5_status.active_display + 1;
	_ts5_status.active_display = display - 1;
	_ts5_data.target = _ts5_data.display_buffer[_ts5_status.active_display];
	al_set_target_bitmap(_ts5_data.target);

	_ts5_status.graphics.target_width = 
		al_get_bitmap_width(_ts5_data.target);
	
	_ts5_status.graphics.target_height = 
		al_get_bitmap_height(_ts5_data.target);

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the active display.
///
/// \return The index of the active display.
///
/// The 'active' display is the display that will be used 
/// for drawing operations.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_active_display()
{
	ts5_check_display("ts5_get_active_display");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_active_display()\n");

	return _ts5_status.active_display + 1;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// @name Display adapter settings
/// A display adapter is a physical monitor that is attached to the computer. 
/// Tscope5 supports multiple display adapters.
/// Functions are available to query the number and dimensions of 
/// each display adapter.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Get the number of display adapters that are connected to the system.
///
/// \return The number of display adapters that are connected to the system.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_num_display_adapters()
{
	ts5_check_display("ts5_get_num_display_adapters");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_num_display_adapters()\n");

	return _ts5_status.num_display_adapters;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the size of the monitor attached to a display adapter.
///
/// \param adapter Index of the display adapter.
/// \param w Variable that will store the width.
/// \param h Variable that will store the height.
///
/// w and h are the maximum size of a window that can be opened on that adapter.
///
/// Pass 0 to get the size of the display adapter for the next display.
///
/// You can pass NULL for values you are not interested in.
////////////////////////////////////////////////////////////////////////////////
void ts5_get_display_adapter_size(const int adapter, double *w, double *h)
{
	ts5_check_display("ts5_get_display_adapter_size");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_adapter_size(%d,%f,%f)\n", 
				adapter, w, h);

	double ww, hh;

	if (adapter < 0) {
		ts5_fatal("ts5_get_display_adapter_size: adapter index not valid\n");
	}
	else if (adapter == 0) {
		ww = _ts5_status.display_adapter[ts5_nextdisplay.adapter].w;
		hh = _ts5_status.display_adapter[ts5_nextdisplay.adapter].h;
	}
	else if (adapter <= _ts5_status.num_display_adapters) {
		ww = _ts5_status.display_adapter[adapter-1].w;
		hh = _ts5_status.display_adapter[adapter-1].h;
	}
	else {
		ts5_fatal("ts5_get_display_adapter_size: adapter index not valid\n");
	}

	if (w) {
		*w = ww;
	}

	if (h) {
		*h = hh;
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Get the width of the monitor attached to a display adapter.
///
/// \param adapter Index of the display adapter.
///
/// \return the width of the display adapter.
///
/// Pass 0 to get the width of the display adapter for the next display.
////////////////////////////////////////////////////////////////////////////////
double ts5_get_display_adapter_width(const int adapter)
{
	ts5_check_display("ts5_get_display_adapter_width");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_adapter_width(%d)\n", adapter);

	double w;
	ts5_get_display_adapter_size(adapter, &w, NULL);

	return w;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the heigth of the monitor attached to a display adapter.
///
/// \param adapter Index of the display adapter.
///
/// \return the height of the display adapter.
///
/// Pass 0 to get the height of the display adapter for the next display.
////////////////////////////////////////////////////////////////////////////////
double ts5_get_display_adapter_height(const int adapter)
{
	ts5_check_display("ts5_get_display_adapter_height");
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_adapter_height(%d)\n", adapter);

	double h;
	ts5_get_display_adapter_size(adapter, NULL, &h);

	return h;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// @name Display settings
/// Before opening a display various settings can be requested 
/// using the functions below.
/// Once a display is opened the settings remain until the end of the program.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Set the display adapter for the next display that will be opened.
///
/// \param adapter Index of the active display adapter.
///
/// \return The index of the previous active display adapter.
///
/// A display adapter corresponds to a physical display that is 
/// attached to the computer.
/// The first adapter has index 1, etc.
///
/// The display is attached to adapter 1 by default.
////////////////////////////////////////////////////////////////////////////////
int ts5_set_display_adapter(const int adapter)
{
	ts5_check_display("ts5_set_display_adapter");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_display_adapter(%d)\n", adapter);

	int retval = ts5_nextdisplay.adapter + 1;

	if (adapter > 0 && adapter <= _ts5_status.num_display_adapters) {
		ts5_nextdisplay.adapter = adapter - 1;
		ts5_log(TS5_LOGLEVEL_4, "%s: %s %d (was %d)\n", 
					"ts5_set_display_adapter", 
					"set display adapter for next display to",
					retval-1, ts5_nextdisplay.adapter+1);
	}
	else {
		ts5_fatal ("ts5_set_display_adapter: adapter index not valid\n");
	}

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the index of the display adapter of a display.
///
/// \param display Index of the display.
///
/// \return The index of the adapter that the display is attached to.
///
/// Pass 0 to get the index of the display adapter for the next display.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_display_adapter(const int display)
{
	ts5_check_display2("ts5_get_display_adapter", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_adapter(%d)\n", display);

	int retval;

	if (display == 0) {
		retval = ts5_nextdisplay.adapter + 1;
	}
	else {
		retval = _ts5_status.display[display-1].adapter + 1;
	}

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Set the size of the next display.
///
/// \param w Width of the next display in pixels.
/// \param h Height of the next display in pixels.
///
/// The default width and height is 320 x 240. This is also the minimum.
////////////////////////////////////////////////////////////////////////////////
void ts5_set_display_size(double w, double h)
{
	ts5_check_display("ts5_set_display_size");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_display_size(%f,%f)\n", w, h);

	int oldw = ts5_nextdisplay.w;
	int oldh = ts5_nextdisplay.h;

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {
		w *= _ts5_status.display_adapter[ts5_nextdisplay.adapter].w;
		h *= _ts5_status.display_adapter[ts5_nextdisplay.adapter].h;
	}

	ts5_nextdisplay.w = w;
	ts5_nextdisplay.h = h;

	ts5_log(TS5_LOGLEVEL_4, "%s: %s %d x %d (was %d x %d)\n", 
				"ts5_set_display_size",
				"set display size for next display to",
				ts5_nextdisplay.w, ts5_nextdisplay.h, oldw, oldh);
}


////////////////////////////////////////////////////////////////////////////////
/// Get the size of a display.
///
/// \param display Index of the display.
/// \param w Variable that will store the width.
/// \param h Variable that will store the height.
///
/// Set display to 0 to get the width of the next display.
///
/// You can pass NULL for values you are not interested in.
////////////////////////////////////////////////////////////////////////////////
void ts5_get_display_size(const int display, double *w, double *h)
{
	ts5_check_display2("ts5_get_display_size", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_size(%d,%p,%p)\n", display, w, h);

	double ww, hh;

	if (display == 0) {
		ww = ts5_nextdisplay.w;
		hh = ts5_nextdisplay.h;
	}
	else {
		ww = _ts5_status.display[display-1].w;
		hh = _ts5_status.display[display-1].h;
	}

	if (w) {
		*w = ww;
	}

	if (h) {
		*h = hh;
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Get the width of a display.
///
/// \param display Index of the display.
///
/// \return the width of the display.
///
/// Set display to 0 to get the width of the next display.
////////////////////////////////////////////////////////////////////////////////
double ts5_get_display_width(const int display)
{
	ts5_check_display2("ts5_get_display_width", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_width(%d)\n", display);

	double w;
	ts5_get_display_size(display, &w, NULL);

	return w;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the height of a display.
///
/// \param display Index of the display.
///
/// \return the height of the display.
///
/// Set display to 0 to get the height of the next display.
////////////////////////////////////////////////////////////////////////////////
double ts5_get_display_height(const int display)
{
	ts5_check_display2("ts5_get_display_height", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_height(%d)\n", display);

	double h;
	ts5_get_display_size(display, NULL, &h);

	return h;
}


////////////////////////////////////////////////////////////////////////////////
/// Set the position of the next display.
///
/// \param x Horizontal position of the display on the desktop.
/// \param y Vertical position of the display on the desktop.
///
/// Only meaningful for non-fullscreen displays.
///
/// CARTESIAN coordinates: Center of the display is the origin, 
/// default position is the center of the display.
///
/// COMPUTER coordinates: The upper left corner of the display is the origin,
/// default position is top left corner of the display.
////////////////////////////////////////////////////////////////////////////////
void ts5_set_display_position(double x, double y)
{
	ts5_check_display("ts5_set_display_position");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_display_position(%f,%f)\n", x, y);

	int oldx = ts5_nextdisplay.x;
	int oldy = ts5_nextdisplay.y;

	if (_ts5_status.graphics.coordinate_scale 
			== TS5_RELATIVE_COORDINATES) {

		if (_ts5_status.graphics.coordinate_system
				== TS5_CARTESIAN_COORDINATES) {
					
			x *= _ts5_status.display_adapter[ts5_nextdisplay.adapter].w
					/ 2.0;
			
			y *= _ts5_status.display_adapter[ts5_nextdisplay.adapter].h
					/ 2.0;
		}
		else {
			x *= _ts5_status.display_adapter[ts5_nextdisplay.adapter].w;
			y *= _ts5_status.display_adapter[ts5_nextdisplay.adapter].h;
		}
	}

	if (_ts5_status.graphics.coordinate_system 
			== TS5_CARTESIAN_COORDINATES) {

		x += _ts5_status.display_adapter[ts5_nextdisplay.adapter].w/2.0;
		x -= ts5_nextdisplay.w/2.0;
		y = -y;
		y += _ts5_status.display_adapter[ts5_nextdisplay.adapter].h/2.0;
		y -= ts5_nextdisplay.h/2.0;
	}

	x += _ts5_status.display_adapter[ts5_nextdisplay.adapter].x1;
	y += _ts5_status.display_adapter[ts5_nextdisplay.adapter].y1;

	ts5_nextdisplay.x = x;
	ts5_nextdisplay.y = y;

	ts5_log(TS5_LOGLEVEL_4, "%s: %s (%d,%d) (was (%d,%d))\n", 
								"ts5_set_display_position",
								"set display position for next display to",
								ts5_nextdisplay.x, ts5_nextdisplay.y, 
								oldx, oldy);
}


////////////////////////////////////////////////////////////////////////////////
/// Get the position of a display.
///
/// \param display Index of the display.
/// \param x Variable that will store the horizontal position.
/// \param y Variable that will store the vertical position.
///
/// Only meaningful for non-fullscreen displays.
///
/// Set display to 0 to get the x position of the next display.
///
/// You can pass NULL for values you are not interested in.
////////////////////////////////////////////////////////////////////////////////
void ts5_get_display_position(const int display, double *x, double *y)
{
	ts5_check_display2("ts5_get_display_position", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_position(%d,%p,%p)\n", 
				display, x, y);

	double xx, yy;
	double dw, dh, aw, ah;

	// get the values we need
	if (display != 0) {

		int adapter = ts5_get_display_adapter(display);
		al_get_window_position(_ts5_data.display[display-1],
								&_ts5_status.display[display-1].x,
								&_ts5_status.display[display-1].y);

		xx = _ts5_status.display[display-1].x;
		yy = _ts5_status.display[display-1].y;

		xx -= _ts5_status.display_adapter[adapter-1].x1;
		yy -= _ts5_status.display_adapter[adapter-1].y1;

		dw = _ts5_status.display[display-1].w;
		dh = _ts5_status.display[display-1].h;

		aw = _ts5_status.display_adapter[adapter-1].w;
		ah = _ts5_status.display_adapter[adapter-1].h;
	}
	else {

		xx = ts5_nextdisplay.x;
		yy = ts5_nextdisplay.y;

		dw = ts5_nextdisplay.w;
		dh = ts5_nextdisplay.h;

		aw = _ts5_status.display_adapter[ts5_nextdisplay.adapter].w;
		ah = _ts5_status.display_adapter[ts5_nextdisplay.adapter].h;
	}

	// if the settings are not default, compute xx and yy
	if (xx!=-1 && yy!=-1) {


		// convert to cartesian if necessary
		if (_ts5_status.graphics.coordinate_system 
				== TS5_CARTESIAN_COORDINATES) {
			xx = xx + dw/2.0 - aw/2.0;
			yy = ah/2.0 - yy - dh/2.0;
		}

		// convert to relative if necessary
		if (_ts5_status.graphics.coordinate_scale 
				== TS5_RELATIVE_COORDINATES) {
						
			if (_ts5_status.graphics.coordinate_system 
					== TS5_CARTESIAN_COORDINATES) {
				xx /= aw/2.0;
				yy /= ah/2.0;
			}
			else {
				xx /= aw;
				yy /= ah;
			}
		}
	}

	if (x) {
		*x = xx;
	}

	if (y) {
		*y = yy;
	}
	
	
}


////////////////////////////////////////////////////////////////////////////////
/// Get the horizontal position of a display.
///
/// \param display Index of the display.
///
/// \return
/// The horizontal position of the display.
////////////////////////////////////////////////////////////////////////////////
double ts5_get_display_x(const int display)
{
	ts5_check_display2("ts5_get_display_x", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_x(%d)\n", display);

	double x;
	ts5_get_display_position(display, &x, NULL);

	return x;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the vertical position of a display.
///
/// \param display Index of the display.
///
/// \return
/// The vertical position of the display.
////////////////////////////////////////////////////////////////////////////////
double ts5_get_display_y(const int display)
{
	ts5_check_display2("ts5_get_display_y", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_y(%d)\n", display);

	double y;
	ts5_get_display_position(display, NULL, &y);

	return y;
}




////////////////////////////////////////////////////////////////////////////////
/// Set the requested refreshrate for the next display.
///
/// \param rate Refreshrate in cycles per second.
///
/// \return The previous refresh rate.
///
/// The refreshrate can only be changed for real fullscreen displays.
///
/// Even then this is only a request.
///
/// The default value 0 (let the system choose).
////////////////////////////////////////////////////////////////////////////////
int ts5_set_refreshrate(const int rate)
{
	ts5_check_display("ts5_set_refreshrate");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_refreshrate(%d)\n", rate);

	int retval = ts5_nextdisplay.refreshrate;

	ts5_nextdisplay.refreshrate = rate;

	ts5_log(TS5_LOGLEVEL_4, "%s: %s %d (was %d)\n", "ts5_set_refreshrate",
			 	"set refreshrate for next display to",
				ts5_nextdisplay.refreshrate, retval);

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the refreshrate of a display.
///
/// \param display Index of the display.
///
/// \return The refresh rate of the display.
///
/// Set display to 0 to get the refreshrate of the next display.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_refreshrate(const int display)
{
	ts5_check_display2("ts5_get_refreshrate", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_refreshrate(%d)\n", display);

	int retval;

	if (display == 0) {
		retval = ts5_nextdisplay.refreshrate;
	}
	else {
		retval = _ts5_status.display[display-1].refreshrate;
	}

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Set the display mode for the next display.
///
/// \param mode Window mode. 
/// Can be TS5_WINDOWED, TS5_FULLSCREEN or TS5_FULLSCREEN_WINDOW.
///
/// \return The previous display mode.
///
/// In TS5_WINDOWED mode the window can have any dimension that 
/// will fit on the physical display 
/// and the position of the window on the display can be adjusted. 
/// Multiple TS5_WINDOWED displays can be combined on one physical display. 
/// The refreshrate is always the refreshrate of the physical display.
///
/// In TS5_FULLSCREEN_WINDOW mode the display always has the dimension 
/// of the desktop. 
/// The refreshrate is always the refreshrate of the physical display. 
/// Only one TS5_FULLSCREEN_WINDOW display can be used
/// per physical display.
///
/// TS5_FULLSCREEN modes are the most flexible: 
/// any combination of display dimension and refreshrate
/// that is supported by the monitor can be set. 
/// Does not seem to work on Mac OS X.
///
/// The default window mode is TS5_WINDOWED.
////////////////////////////////////////////////////////////////////////////////
int ts5_set_display_mode(const int mode)
{
	ts5_check_display("ts5_set_display_mode");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_display_mode(%d)\n", mode);

	int retval = ts5_nextdisplay.display_mode;

	if (mode != TS5_WINDOWED 
			&& mode != TS5_FULLSCREEN 
			&& mode != TS5_FULLSCREEN_WINDOW) {
		ts5_fatal ("%s: %s (%d)\n", "ts5_set_display_mode",
				 		"requested display mode not available", mode);
	}

	ts5_nextdisplay.display_mode = mode;

	ts5_log(TS5_LOGLEVEL_4, "%s: %s %d (was %d)\n", "ts5_set_display_mode",
					"set display mode for next display to",
					ts5_nextdisplay.display_mode, retval);

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the display mode of a display.
///
/// \param display Index of the display.
///
/// \return The window mode of the display.
///
/// Set display to 0 to get the display mode of the next display.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_display_mode(const int display)
{
	ts5_check_display2("ts5_get_display_mode", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_display_mode(%d)\n", display);

	int retval;

	if (display == 0) {
		retval = ts5_nextdisplay.display_mode;
	}
	else {
		retval = _ts5_status.display[display-1].display_mode;
	}

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Set the vsync mode for the next display.
///
/// \param mode Vsync mode. 
/// Can be TS5_VSYNC_WHATEVER, TS5_VSYNC_ON or TS5_VSYNC_OFF.
///
/// \return The previous vsync mode.
///
/// The default vsync mode is TS5_VSYNC_ON.
////////////////////////////////////////////////////////////////////////////////
int ts5_set_vsync_mode(const int mode)
{
	ts5_check_display("ts5_set_vsync_mode");
	ts5_log(TS5_LOGLEVEL_4, "ts5_set_vsync_mode(%d)\n", mode);

	int retval = ts5_nextdisplay.vsync_mode;

	if (mode < TS5_VSYNC_WHATEVER || mode > TS5_VSYNC_OFF) {
		ts5_fatal("ts5_set_vsync_mode: mode should be 0, 1 or 2 (is %d)\n", 
					mode);
	}

	ts5_nextdisplay.vsync_mode = mode;

	ts5_log(TS5_LOGLEVEL_4, "%s: %s %d (was %d)\n", "ts5_set_vsync_mode",
			 		"set vsync mode for next display to",
					ts5_nextdisplay.vsync_mode, retval);

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Get the vsync mode of a display.
///
/// \param display Index of the display.
///
/// \return The vsync mode of the display.
///
/// Set display to 0 to get the requested vsync mode for the next display.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_vsync_mode(const int display)
{
	ts5_check_display2("ts5_get_vsync_mode", display);
	ts5_log(TS5_LOGLEVEL_4, "ts5_get_vsync_mode(%d)\n", display);

	int retval;

	if (display == 0) {
		retval = ts5_nextdisplay.vsync_mode;
	}
	else {
		retval = _ts5_status.display[display-1].vsync_mode;
	}

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////
