////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file parport_internal.c
/// Definitions of internal parallel port functions.
////////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/parport_internal.h"
#include "../include/tscope5/system_internal.h"
#include "../include/tscope5/timer_internal.h"

/// Is the parport subsystem installed?
int _ts5_is_parport_installed = 0;

#ifdef TS5_WINDOWS
	#include <windows.h>
	HINSTANCE hLib;
	typedef short (_stdcall *inbptr)(short portaddr);
	typedef short (_stdcall *outbptr)(short portaddr, short value);
	inbptr inb;
	outbptr outb;

#elif defined TS5_LINUX
	#include <unistd.h>
	#include <sys/io.h>

#endif


////////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each parport function.
///
/// \param calling_function Name the function that calls for 
///	this check or installation.
///
/// Checks whether Tscope5 is installed. If not, Tscope5 is installed.
/// Then checks whether the parport subsystem is installed. 
/// If so, aborts with an error message.
////////////////////////////////////////////////////////////////////////////////
void ts5_check_parport(char *calling_function)
{
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_parport\n", calling_function);

	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5(calling_function);
	}
	
	if (_ts5_is_parport_installed) {
		ts5_fatal("%s: parport subsystem is already running. ",
					"This function has to be called before the (automatic) ",
					"installation of the parallel port system.\n",
					calling_function);
	}	
}
	
			

////////////////////////////////////////////////////////////////////////////////
/// Do some checks at the start of each parport function.
///
/// \param calling_function Name the function that calls for 
///	this check or installation.
///
/// Checks whether the parport subsystem is installed. 
/// If not, the parport subsystem is installed.
////////////////////////////////////////////////////////////////////////////////
void ts5_check_parport2(char *calling_function)
{
	ts5_log(TS5_LOGLEVEL_6, "%s: ts5_check_parport2\n", calling_function);

	if (!_ts5_is_parport_installed) {
		ts5_install_parport(calling_function);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Install the parport subsystem.
///
/// \param calling_function Name the function that calls for
/// this check or installation.
///
/// This function is called automatically if necessary.
////////////////////////////////////////////////////////////////////////////////
void ts5_install_parport(char *calling_function)
{
	if (!_ts5_is_timer_installed) {
		ts5_install_timer(calling_function);
	}

	if (!_ts5_is_parport_installed) {

		_ts5_is_parport_installed = 1;
		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5 parport\n", 
				calling_function);

		int i;
				
		#ifdef TS5_MACOSX
			ts5_fatal("%s: parallel port boxes not supported on Mac OS X\n", 
				calling_function);

		#endif

		#ifdef TS5_WINDOWS		
			if (!(hLib = LoadLibrary("inpout32.dll"))) {
				ts5_fatal("%s: could not load inpout32.dll\n", 
						calling_function);
			}

			if (!(inb = (inbptr)GetProcAddress(hLib, "Inp32"))) {
				ts5_fatal("%s: could not get address for Inp32\n", 
						calling_function);
			}

			if (!(outb = (outbptr)GetProcAddress(hLib, "Out32"))) {
				ts5_fatal("%s: could not get address for Out32\n", 
						calling_function);
			}

			for (i=0; i<_ts5_status.timer.num_parports; i++) {
				outb(_ts5_status.timer.parport_control_register[i], 0);
			}

		#endif

		#ifdef TS5_LINUX
			if (geteuid() != 0) {
				ts5_fatal("%s: only root can open the parallel port.", 
						calling_function);
			}

			if (ioperm(TS5_PARPORT_1_DATA, 3, 1)) {
				ts5_fatal("%s: error opening parallell port 1.", 
						calling_function );
			}

			if (ioperm(TS5_PARPORT_2_DATA, 3, 1)) {
				ts5_fatal("%s: error opening parallell port 2.", 
						calling_function);
			}

			if (ioperm(TS5_PARPORT_3_DATA, 3, 1)) {
				ts5_fatal("%s: error opening parallell port 3.", 
						calling_function);
			}

			for (i=0; i<_ts5_status.timer.num_parports; i++) {
				outb(0, _ts5_status.timer.parport_control_register[i]);
			}

		#endif

		// unitialize event source
		al_init_user_event_source(
			&_ts5_data.timer.parport_response_event_source);

		al_init_user_event_source(
			&_ts5_data.timer.parport_trigger_event_source);

		// register event source
		al_register_event_source(_ts5_data.timer.response_queue,
			&_ts5_data.timer.parport_response_event_source);

		al_register_event_source(_ts5_data.timer.trigger_queue,
			&_ts5_data.timer.parport_trigger_event_source);

		al_register_event_source(_ts5_data.timer.trigger_log,
			&_ts5_data.timer.parport_trigger_event_source);

		// initialize thread for parport
		_ts5_data.timer.parport_thread = 
			al_create_thread(ts5_parport_threadfunc, NULL);
		
		al_start_thread(_ts5_data.timer.parport_thread);

		// assign response codes
		TS5_TIMER_STATUS *timer = &_ts5_status.timer;
		
		timer->num_parports = 3;
		timer->parport =
			(TS5_PARPORT_STATUS *)al_malloc(sizeof(TS5_PARPORT_STATUS)*
			timer->num_parports);

		for (i=0; i<timer->num_parports; i++) {

			timer->parport[i].num_buttons = 5;
			timer->parport[i].num_defined_buttons = 0;
			timer->parport[i].num_active_buttons = 0;
			timer->parport[i].button_debounce_time = 0.0;

			timer->parport[i].status_register_current_state =
				(int *)al_malloc(sizeof(int) 
				* timer->parport[i].num_buttons);

			timer->parport[i].status_register_previous_state =
				(int *)al_malloc(sizeof(int)
				* timer->parport[i].num_buttons);

			timer->parport[i].status_register_last_state_toggle_time =
				(double *)al_malloc(sizeof(double)
				* timer->parport[i].num_buttons);

			timer->parport[i].button_press_defined =
				(int *)al_malloc(sizeof(int)
				* timer->parport[i].num_buttons);

			timer->parport[i].button_press_active =
				(int *)al_malloc(sizeof(int)
				* timer->parport[i].num_buttons);

			timer->parport[i].button_release_defined =
				(int *)al_malloc(sizeof(int)
				* timer->parport[i].num_buttons);

			timer->parport[i].button_release_active =
				(int *)al_malloc(sizeof(int)
				* timer->parport[i].num_buttons);

			int j;
			for (j=0; j<timer->parport[i].num_buttons; j++) {
				timer->parport[i].status_register_current_state[j] = 0;
				timer->parport[i].status_register_previous_state[j] = 0;

				timer->parport[i].status_register_last_state_toggle_time[j] = 
					al_get_time();

				timer->parport[i].button_press_defined[j] = 0;
				timer->parport[i].button_press_active[j] = 0;
				timer->parport[i].button_release_defined[j] = 0;
				timer->parport[i].button_release_active[j] = 0;
			}

			timer->parport[i].is_trigger_input_device = 0;
			timer->parport[i].is_trigger_output_device = 0;
			timer->parport[i].trigger_debounce_time = 0.0;

			timer->parport[i].simulate_trigger_input = 0;
			timer->parport[i].trigger_simulation_interval = 2.0;

			timer->parport[i].last_trigger_simulation = 
				al_get_time();

			timer->parport[i].data_register_current_state = 0;
			timer->parport[i].data_register_previous_state = 0;
			timer->parport[i].trigger_output_time = 0.002;
		}

		atexit(ts5_uninstall_parport);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Uninstall the parport subsystem.
///
/// This function is called automatically at the end of the program.
////////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_parport()
{
	if (_ts5_is_parport_installed) {

		int i;
		
		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5 parport\n");

		// remove thread
		al_join_thread(_ts5_data.timer.parport_thread, NULL);
		al_destroy_thread(_ts5_data.timer.parport_thread);

		// unregister event source
		al_unregister_event_source(_ts5_data.timer.response_queue,
			&_ts5_data.timer.parport_response_event_source);

		al_unregister_event_source(_ts5_data.timer.trigger_queue,
			&_ts5_data.timer.parport_trigger_event_source);

		al_unregister_event_source(_ts5_data.timer.trigger_log,
			&_ts5_data.timer.parport_trigger_event_source);

		// remove event source
		al_destroy_user_event_source(
			&_ts5_data.timer.parport_response_event_source);
		
		al_destroy_user_event_source(
			&_ts5_data.timer.parport_trigger_event_source);

		#ifdef TS5_WINDOWS
			for (i=0; i<_ts5_status.timer.num_parports; i++) {
				outb(_ts5_status.timer.parport_control_register[i], 0);
			}

			FreeLibrary(hLib);

		#endif

		#ifdef TS5_LINUX
			for (i=0; i<_ts5_status.timer.num_parports; i++) {
				outb(0, _ts5_status.timer.parport_control_register[i]);
			}

			if (geteuid() != 0) {
				ts5_fatal("%s: %s\n",
						"ts5_uninstall_parport", 
						"only root can close the parallel port");
			}

			if (ioperm(TS5_PARPORT_1_DATA, 3, 0)) {
				ts5_fatal("%s: %s\n",
						"ts5_uninstall_parport", 
						"error closing parallell port 1");
			}

			if (ioperm(TS5_PARPORT_2_DATA, 3, 0)) {
				ts5_fatal("%s: %s\n",
						"ts5_uninstall_parport", 
						"error closing parallell port 2");
			}

			if (ioperm(TS5_PARPORT_3_DATA, 3, 0)) {
				ts5_fatal("%s: %s\n",
						"ts5_uninstall_parport", 
						"error closing parallell port 3");
			}

		#endif

		TS5_TIMER_STATUS *timer = &_ts5_status.timer;

		for (i=0; i<timer->num_parports; i++) {

			al_free(timer->parport[i].status_register_current_state);
			timer->parport[i].status_register_current_state = NULL;

			al_free(timer->parport[i].status_register_previous_state);
			timer->parport[i].status_register_previous_state = NULL;

			al_free(timer->parport[i].status_register_last_state_toggle_time);
			timer->parport[i].status_register_last_state_toggle_time = NULL;

			al_free(timer->parport[i].button_press_defined);
			timer->parport[i].button_press_defined = NULL;

			al_free(timer->parport[i].button_press_active);
			timer->parport[i].button_press_active = NULL;

			al_free(timer->parport[i].button_release_defined);
			timer->parport[i].button_release_defined = NULL;

			al_free(timer->parport[i].button_release_active);
			timer->parport[i].button_release_active = NULL;
		}

		al_free(timer->parport);
		timer->parport=NULL;

		timer->num_parports = 0;
		timer->parport_is_response_device = 0;

		_ts5_is_parport_installed = 0;
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Set the DATA register of the parallel port to input mode.
///
/// \param calling_function Name the function that calls 
/// for this check or installation.
/// \param device Number of the parallel port.
////////////////////////////////////////////////////////////////////////////////
void ts5_set_parport_trigger_input(char *calling_function, int device)
{
	ts5_check_parport2(calling_function);

	TS5_TIMER_STATUS *timer = &_ts5_status.timer;

	if (!timer->parport[device-1].is_trigger_input_device) {

		if (device<1 || device>timer->num_parports) {
			ts5_fatal("%s: %s %d, %s %d\n",
						calling_function, 
						"device argument is", device, 
						"number of parallel ports is", timer->num_parports);
		}

		if (timer->parport[device-1].num_defined_buttons) {
			ts5_fatal("%s: %s %d %s\n", calling_function, "parallel port", 
					device, "is already defined as an response device");
		}

		if (timer->parport[device-1].is_trigger_output_device) {
			ts5_fatal("%s: %s %d %s\n", calling_function, "parallel port", 
					device, "is already defined as a trigger output device");
		}

		timer->parport_is_trigger_device = 1;
		timer->parport[device-1].is_trigger_input_device = 1;

		#ifdef TS5_WINDOWS
			outb(_ts5_status.timer.parport_control_register[device-1], 32);

		#elif defined TS5_LINUX
			outb(32, _ts5_status.timer.parport_control_register[device-1]);

		#endif

	}
}


////////////////////////////////////////////////////////////////////////////////
/// Set the DATA register of the parallel port to output mode.
///
/// \param calling_function Name the function that calls 
/// for this check or installation.
/// \param device Number of the parallel port.
////////////////////////////////////////////////////////////////////////////////
void ts5_set_parport_trigger_output(char *calling_function, int device)
{
	ts5_check_parport2(calling_function);
	TS5_TIMER_STATUS *timer = &_ts5_status.timer;

	if (!timer->parport[device-1].is_trigger_output_device) {

		if (device<1 || device>timer->num_parports) {
			ts5_fatal("%s: %s %d, %s %d\n", calling_function, 
					"device argument is", device, 
					"number of parallel ports is", timer->num_parports);
		}

		if (timer->parport[device-1].num_defined_buttons) {
			ts5_fatal("%s: %s %d %s\n", calling_function, "parallel port", 
					device, "is already defined as an response device");
		}

		if (timer->parport[device-1].is_trigger_input_device) {
			ts5_fatal("%s: %s %d %s\n", calling_function, "parallel port", 
					device, "is already defined as a trigger input device");
		}

		timer->parport_is_trigger_device = 1;
		timer->parport[device-1].is_trigger_output_device = 1;

		#ifdef TS5_WINDOWS
			outb(_ts5_status.timer.parport_control_register[device-1], 0);

		#elif defined TS5_LINUX
			outb(0, _ts5_status.timer.parport_control_register[device-1]);

		#endif

	}
}


////////////////////////////////////////////////////////////////////////////////
/// Parallel port thread function.
///
/// \warning This is an internal function, no checks are performed.
////////////////////////////////////////////////////////////////////////////////
void *ts5_parport_threadfunc()
{
	ALLEGRO_EVENT parport_event;
	double previous_time;
	double current_time = al_get_time();
	
	TS5_TIMER_STATUS *timer = &_ts5_status.timer;
	
	while (!al_get_thread_should_stop(_ts5_data.timer.parport_thread)) {

		int i;
		for (i=0; i<timer->num_parports; i++) {


			// monitor STATUS register for responses
			short input = 0;

			#ifndef TS5_MACOSX
				input = inb(_ts5_status.timer.parport_status_register[i]);

			#endif

			timer->parport[i].status_register_current_state[0] = 
					(input &   8) >> 3;
			
			timer->parport[i].status_register_current_state[1] = 
					(input &  16) >> 4;
			
			timer->parport[i].status_register_current_state[2] = 
					(input &  32) >> 5;
			
			timer->parport[i].status_register_current_state[3] = 
					(input &  64) >> 6;
			
			timer->parport[i].status_register_current_state[4] = 
					(input & 128) >> 7;

			int button = 0;
			int action = 0;

			int j;
			for (j=0; j<timer->parport[i].num_buttons; j++) {

				if (timer->parport[i].status_register_current_state[j] !=
					timer->parport[i].status_register_previous_state[j]) {

					if (current_time - timer->parport[i]
							.status_register_last_state_toggle_time[j] >
							timer->parport[i].button_debounce_time) {

						button = j+1;
						action = timer->parport[i]
									.status_register_current_state[j];

						timer->parport[i].status_register_previous_state[j] =
							timer->parport[i].status_register_current_state[j];

						timer->parport[i]
							.status_register_last_state_toggle_time[j] = current_time;
					}
				}
			}

			if (button) {

				if (action==0) {
					parport_event.user.type = TS5_EVENT_PARPORT_BUTTON_UP;
				}

				if (action==1) {
					parport_event.user.type = TS5_EVENT_PARPORT_BUTTON_DOWN;
				}

				parport_event.user.data1 = i;
				parport_event.user.data2 = button;
				parport_event.user.data3 =  
					(intptr_t)((current_time-previous_time)*1000000.0);

				al_emit_user_event(
					&_ts5_data.timer.parport_response_event_source,
					&parport_event, NULL);
			}


			// monitor DATA register for triggers
			input = 0;

			#ifndef TS5_MACOSX	
				input = inb(_ts5_status.timer.parport_data_register[i]);

			#endif

			timer->parport[i].data_register_current_state = input;

			if (timer->parport[i].is_trigger_input_device &&
					timer->parport[i].data_register_current_state &&
					timer->parport[i].data_register_current_state !=
					timer->parport[i].data_register_previous_state) {

				if (current_time 
						- timer->parport[i].data_register_last_state_toggle_time
						> timer->parport[i].trigger_debounce_time) {

					parport_event.user.type = TS5_EVENT_PARPORT_TRIGGER;

					parport_event.user.data1 = i;
					
					parport_event.user.data2 = 
						(int)timer->parport[i].data_register_current_state;
					
					parport_event.user.data3 = 
						(intptr_t)((current_time-previous_time)*1000000.0);

					al_emit_user_event(
						&_ts5_data.timer.parport_trigger_event_source,
						&parport_event, NULL);

					timer->parport[i].data_register_previous_state =
						timer->parport[i].data_register_current_state;

					timer->parport[i].data_register_last_state_toggle_time = 
						current_time;
				}
			}

			// simulate triggers
			if (timer->parport[i].simulate_trigger_input) {

				if (current_time - timer->parport[i].last_trigger_simulation
						>= timer->parport[i].trigger_simulation_interval) {

					parport_event.user.type = TS5_EVENT_PARPORT_TRIGGER;

					parport_event.user.data1 = i;

					parport_event.user.data2 = 
						timer->parport[i].simulate_trigger_input;

					parport_event.user.data3 = 
						(intptr_t)((current_time-previous_time)*1000000.0);

					al_emit_user_event(
						&_ts5_data.timer.parport_trigger_event_source,
						&parport_event, NULL);

					timer->parport[i].last_trigger_simulation = current_time;
				}
			}
		}
		previous_time = current_time;
		current_time = al_get_time();
	}

	return NULL;
}


