////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file system.c
/// Definitions of system functions.
/// \example system01.c
/// \example system02.c
////////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/system.h"
#include "../include/tscope5/display.h"
#include "../include/tscope5/system_internal.h"


////////////////////////////////////////////////////////////////////////////////
/// @name Logging functions
/// Tscope5 automatically sends information about its status
/// to a logging window and a file called tscope5.log.
///
/// With the functions below you control the amount of information
/// that is logged.
/// It is also possible to send your own log messages.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Send info to a logging window.
///
/// \param level The importance of this message
/// (TS5_LOGLEVEL_0 - TS5_LOGLEVEL_6). Smaller values are more important.
/// \param format Printf-style format string that contains the log message.
///
/// The log message is printed in a separate text log window.
///
/// Messages with a level above the level specified
/// with ts5_set_log_level are omitted.
////////////////////////////////////////////////////////////////////////////////
void ts5_log(const unsigned int level, const char *format, ...)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5("ts5_log");
	}

	if (level <= _ts5_status.loglevel) {

		if (!_ts5_is_textlog_installed) {
			_ts5_data.textlog = fopen("tscope5.log", "w+");
			if (!_ts5_data.textlog) {
				ts5_fatal("ts5_log: could not open log file\n");
			}
			_ts5_is_textlog_installed = 1;
		}

		char outbuf[TS5_MAX_CHAR];
		va_list args;
		va_start(args, format);
		vsnprintf(outbuf, TS5_MAX_CHAR*sizeof(char), format, args);
		va_end(args);

		double logtime = al_get_time();
		fprintf(stdout, "%8.3f: %s", logtime, outbuf);
		//fflush(stdout);
		fprintf(_ts5_data.textlog, "%8.3f: %s", logtime, outbuf);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Set log level
///
/// \param loglevel How much stuff is logged.
///
/// \return The previous log level.
///
/// This parameter function controls what information
/// is printed to the log window.
///
/// There are six possible settings:
/// - TS5_LOGLEVEL_0: no debugging info.
/// - TS5_LOGLEVEL_1: report when Tscope5 subsystems are loaded.
/// - TS5_LOGLEVEL_2: report when memory is allocated/freed.
/// - TS5_LOGLEVEL_3: report when response buttons are defined.
/// - TS5_LOGLEVEL_4: report when parameter are set or queried.
/// - TS5_LOGLEVEL_5: report each function call.
/// - TS5_LOGLEVEL_6: report internal check and coordinate functions.
///
/// If other values are passed to this function the program is aborted.
///
/// The default loglevel is TS5_LOGLEVEL_3.
////////////////////////////////////////////////////////////////////////////////
int ts5_set_log_level(const unsigned int loglevel)
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5("ts5_set_log_level");
	}

	ts5_log(TS5_LOGLEVEL_4, "ts5_set_log_level(%d)\n", loglevel);

	int retval;
	retval = _ts5_status.loglevel;

	if (loglevel <= TS5_LOGLEVEL_6) {
		_ts5_status.loglevel = loglevel;

		ts5_log(TS5_LOGLEVEL_4, "%s: %s %d to %d\n", "ts5_set_log_level",
					"changed loglevel from", retval, loglevel);
	}
	else {
		ts5_fatal("ts5_log: requested loglevel not supported.");
	}

	return retval;
}


////////////////////////////////////////////////////////////////////////////////
/// Get log level
///
/// \return The log level.
////////////////////////////////////////////////////////////////////////////////
int ts5_get_log_level()
{
	if (!_ts5_is_tscope5_installed) {
		ts5_install_tscope5("ts5_get_log_level");
	}

	ts5_log(TS5_LOGLEVEL_4, "ts5_get_log_level\n");

	return _ts5_status.loglevel;
}


////////////////////////////////////////////////////////////////////////////////
/// Write the status of all Tscope5 subsystem to the log window.
///
/// Information is printed about:
/// - The display adapter(s) of the computer and the available fullscreen modes.
/// - The displays opened by Tscope5.
/// - The graphics parameters.
////////////////////////////////////////////////////////////////////////////////
void ts5_print_status()
{
	char line[]="==================================================\n";
	int i;

	// display adapters + supported fullscreen modes

	ts5_log(TS5_LOGLEVEL_0, "\n");
	ts5_log(TS5_LOGLEVEL_0, line);
	ts5_log(TS5_LOGLEVEL_0, "tscope5 status:\n");
	ts5_log(TS5_LOGLEVEL_0, "\n");

	ts5_log(TS5_LOGLEVEL_0, "Loglevel: %d\n", _ts5_status.loglevel);
	ts5_log(TS5_LOGLEVEL_0, "Number of display adapters: %d\n",
				_ts5_status.num_display_adapters);

	for (i = 0; i < _ts5_status.num_display_adapters; i++) {

		ts5_log(TS5_LOGLEVEL_0, "%s %d: w: %d (%d:%d), h: %d (%d:%d)\n",
				"Display adapter", i+1,
				_ts5_status.display_adapter[i].w,
				_ts5_status.display_adapter[i].x1,
				_ts5_status.display_adapter[i].x2,
				_ts5_status.display_adapter[i].h,
				_ts5_status.display_adapter[i].y1,
				_ts5_status.display_adapter[i].y2);
	}

	ts5_log(TS5_LOGLEVEL_0, "\n");

	int oldadapter = ts5_nextdisplay.adapter;
	ALLEGRO_DISPLAY_MODE *mode;
	mode = (ALLEGRO_DISPLAY_MODE *)al_malloc(sizeof(ALLEGRO_DISPLAY_MODE));

	for (i = 0; i < _ts5_status.num_display_adapters; i++) {

		al_set_new_display_adapter(i);
		int oldrate = al_get_new_display_refresh_rate();
		int oldflags = al_get_new_display_flags();
		al_set_new_display_refresh_rate(0);
		al_set_new_display_flags(0);

		int nmodes = al_get_num_display_modes();

		int j;
		for (j=0; j<nmodes; j++) {

			mode = al_get_display_mode(j, mode);
			ts5_log(TS5_LOGLEVEL_0,
					"adapter %d mode %2d: w=%4d, h=%4d, format=%d, rate=%d\n",
					i+1, j+1, mode->width, mode->height,
					mode->format, mode->refresh_rate);
		}

		al_set_new_display_refresh_rate(oldrate);
		al_set_new_display_flags(oldflags);
	}

	ts5_log(TS5_LOGLEVEL_0, "\n");
	al_set_new_display_adapter(oldadapter);
	al_free(mode);

	// display settings + settings for next display
	ts5_log(TS5_LOGLEVEL_0, line);
	ts5_log(TS5_LOGLEVEL_0, "Parameters for next display:\n");

	ts5_log(TS5_LOGLEVEL_0, "\tAdapter %d\n",
				ts5_nextdisplay.adapter+1);

	ts5_log(TS5_LOGLEVEL_0, "\tw=%d, h=%d\n",
				ts5_nextdisplay.w, ts5_nextdisplay.h);

	ts5_log(TS5_LOGLEVEL_0, "\tx=%d, y=%d\n",
				ts5_nextdisplay.x, ts5_nextdisplay.y);

	ts5_log(TS5_LOGLEVEL_0, "\tRefreshrate=%d\n",
				ts5_nextdisplay.refreshrate);

	ts5_log(TS5_LOGLEVEL_0, "\tDisplay mode=%d\n",
				ts5_nextdisplay.display_mode);

	ts5_log(TS5_LOGLEVEL_0, "\tVsync mode=%d\n",
				ts5_nextdisplay.vsync_mode);

	ts5_log(TS5_LOGLEVEL_0, "\n");

	ts5_log(TS5_LOGLEVEL_0, "Number of displays: %d\n",
				_ts5_status.num_displays);

	ts5_log(TS5_LOGLEVEL_0, "Active display: %d\n",
				_ts5_status.active_display);

	for (i = 0; i < _ts5_status.num_displays; i++) {

		ts5_log(TS5_LOGLEVEL_0,
				"%s %d: %s %d, w=%d, h=%d, x=%d, y=%d, %s=%d, %s=%d, %s=%d\n",
				"Display", i+1,
				"Adapter", _ts5_status.display[i].adapter+1,
				_ts5_status.display[i].w, _ts5_status.display[i].h,
				_ts5_status.display[i].x, _ts5_status.display[i].y,
				"Refreshrate", _ts5_status.display[i].refreshrate,
				"Dispay mode", _ts5_status.display[i].display_mode,
				"Vsync mode", _ts5_status.display[i].vsync_mode);
	}

	ts5_log(TS5_LOGLEVEL_0, "\n");

	// graphics parameters
	float r, g, b, a;
	ts5_log(TS5_LOGLEVEL_0, line);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
				"Coordinate system", _ts5_status.graphics.coordinate_system);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
				"Coordinate scale", _ts5_status.graphics.coordinate_scale);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d x %d\n",
				"Target bitmap size",
				_ts5_status.graphics.target_width,
				_ts5_status.graphics.target_height);

	al_unmap_rgba_f(_ts5_status.graphics.foreground_color, &r, &g, &b, &a);
	ts5_log(TS5_LOGLEVEL_0, "%s: r: %f  g: %f b: %f alpha: %f\n",
			"Foreground color", r, g, b, a);

	al_unmap_rgba_f(_ts5_status.graphics.background_color, &r, &g, &b, &a);
	ts5_log(TS5_LOGLEVEL_0, "%s: r: %f  g: %f b: %f alpha: %f\n",
			"Background color", r, g, b, a);

	ts5_log(TS5_LOGLEVEL_0, "%s: %f\n",
			"Line drawing thickness", _ts5_status.graphics.drawing_thickness);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
			"Fill mode", _ts5_status.graphics.fill_mode);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
			"Font type", _ts5_status.graphics.font_type);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
			"Font style", _ts5_status.graphics.font_style);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
			"Font size", _ts5_status.graphics.font_size);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
			"Font number", _ts5_status.graphics.font_index);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
			"Text alignment", _ts5_status.graphics.text_alignment);

	ts5_log(TS5_LOGLEVEL_0, "\n");


	// audio parameters
	ts5_log(TS5_LOGLEVEL_0, line);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
			"Number of audio channels", _ts5_status.audio.channels);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
			"Sample rate", _ts5_status.audio.samplerate);

	ts5_log(TS5_LOGLEVEL_0, "%s: %d\n",
			"Sample format", _ts5_status.audio.depth);

	ts5_log(TS5_LOGLEVEL_0, "%s: %f\n",
			"Audio Gain", _ts5_status.audio.gain);


	// timer
	TS5_TIMER_STATUS *timer = &_ts5_status.timer;

	ts5_log(TS5_LOGLEVEL_0, line);

	ts5_log(TS5_LOGLEVEL_0, "Program priority: %+d\n",
			timer->priority);

	ts5_log(TS5_LOGLEVEL_0, "Number of defined response buttons: %d\n",
			timer->num_defined_buttons);

	ts5_log(TS5_LOGLEVEL_0, "Number of active response buttons: %d\n",
			timer->num_active_buttons);

	ts5_log(TS5_LOGLEVEL_0, "Voicekey is response device: %d\n",
			timer->voicekey_is_response_device);
			
	if (timer->voicekey_is_response_device) {

		for (i=0; i<timer->voicekey.num_buttons; i++) {

			if (timer->voicekey.button_press_active[i]!=0) {
				ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d\n",
					"Voice key button press", i+1, "has response value",
					timer->voicekey.button_press_defined[i]);
			}

			if (timer->voicekey.button_release_active[i]!=0) {
				ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d\n",
					"Voice key button release", i+1, "has response value",
					timer->voicekey.button_release_defined[i]);
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "Mouse is response device: %d\n",
			timer->mouse_is_response_device);

	if (timer->mouse_is_response_device) {

		for (i=0; i<timer->mouse.num_buttons; i++) {

			if (timer->mouse.button_press_active[i]!=0) {
				ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d\n",
					"Mouse button press", i+1, "has response value",
					timer->mouse.button_press_defined[i]);
			}

			if (timer->mouse.button_release_active[i]!=0) {
				ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d\n",
					"Mouse button release", i+1, "has response value",
					timer->mouse.button_release_defined[i]);
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "Keyboard is response device: %d\n",
			timer->keyboard_is_response_device);

	if (timer->keyboard_is_response_device) {

		for (i=0; i<timer->keyboard.num_buttons; i++) {

			if (timer->keyboard.button_press_active[i]!=0) {
				ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d\n",
					"Keyboard button press", i+1, "has response value",
					timer->keyboard.button_press_defined[i]);
			}

			if (timer->keyboard.button_release_active[i]!=0) {
				ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d\n",
					"Keyboard button release", i+1, "has response value",
					timer->keyboard.button_release_defined[i]);
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "Number of joysticks: %d\n",
			timer->num_joysticks);

	ts5_log(TS5_LOGLEVEL_0, "Joystick is response device: %d\n",
			timer->joystick_is_response_device);

	if (timer->joystick_is_response_device) {

		for (i=0; i<timer->num_joysticks; i++) {
			int j;

			for (j=0; j<timer->joystick[i].num_buttons; j++) {

				if (timer->joystick[i].button_press_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d %s %d\n",
						"Joystick", i+1, "button press", j+1,
						"has response value",
						timer->joystick[i].button_press_defined[j]);
				}

				if (timer->joystick[i].button_release_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d %s %d\n",
						"Joystick", i+1, "button release", j+1,
						"has response value",
						timer->joystick[i].button_release_defined[j]);
				}
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "Number of cedrusboxes: %d\n",
			timer->num_cedrusboxes);

	ts5_log(TS5_LOGLEVEL_0, "Cedrusbox is response device: %d\n",
			timer->cedrusbox_is_response_device);

	if (timer->cedrusbox_is_response_device) {

		for (i=0; i<timer->num_cedrusboxes; i++) {

			ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d\n",
				"Cedrusbox", i+1,
				"is mapped to serial device",
				timer->cedrusbox[i].port_num+1);

			ts5_log(TS5_LOGLEVEL_0, "Cedrusbox %d is type %d\n",
				i+1, timer->cedrusbox[i].type);

			int j;
			for (j=0; j<timer->cedrusbox[i].num_buttons; j++) {

				if (timer->cedrusbox[i].button_press_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d %s %d\n",
						"Cedrusbox", i+1, "button press", j+1,
						"has response value",
						timer->cedrusbox[i].button_press_defined[j]);
				}

				if (timer->cedrusbox[i].button_release_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d %s %d\n",
						"Cedrusbox", i+1, "button release", j+1,
						"has response value",
						timer->cedrusbox[i].button_release_defined[j]);
				}
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "Number of parallel ports: %d\n",
			timer->num_parports);

	ts5_log(TS5_LOGLEVEL_0, "Parallel port is response device: %d\n",
			timer->parport_is_response_device);

	if (timer->parport_is_response_device) {

		for (i=0; i<timer->num_parports; i++) {

			int j;
			for (j=0; j<timer->parport[i].num_buttons; j++) {

				if (timer->parport[i].button_press_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d %s %d\n",
						"Parallel box", i+1, "button press", j+1,
						"has response value",
						timer->parport[i].button_press_defined[j]);
				}

				if (timer->cedrusbox[i].button_release_active[j]!=0) {
					ts5_log(TS5_LOGLEVEL_0, "%s %d %s %d %s %d\n",
						"Parallel box", i+1, "button release", j+1,
						"has response value",
						timer->parport[i].button_release_defined[j]);
				}
			}
		}
	}

	ts5_log(TS5_LOGLEVEL_0, "\n");



	ts5_log(TS5_LOGLEVEL_0, "\n");
	ts5_log(TS5_LOGLEVEL_0, line);
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// @name Error exit
/// During an experiment lots of unexpected things can happen.
/// (e.g. you forgot to copy stimulus bitmaps to the experiment computer).
///
/// It is a good habit to check for such unexpected events and exit your
/// program safely before an ugly crash happens.
//@{
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
/// Exit safely with an error message.
///
/// \param format Printf-style format string that contains the error message.
///
/// Closes Tscope5, all subsystems and the log window.
///
/// A message box appears with the error message supplied by the user.
///
/// This function can be called before Tscope5 is installed.
////////////////////////////////////////////////////////////////////////////////
void ts5_fatal(const char *format, ...)
{
	ts5_log(TS5_LOGLEVEL_1, "ts5_fatal\n");

	char outbuf[TS5_MAX_CHAR];
	va_list args;
	va_start(args, format);
	vsnprintf(outbuf, TS5_MAX_CHAR*sizeof(char), format, args);
	va_end(args);

	fprintf(stdout, "Tscope5 error: %s", outbuf);

	#ifndef TS5_RASPBERRYPI
		if (_ts5_status.num_displays>0) {
	
			if (_ts5_status.display[_ts5_status.active_display].display_mode
					== TS5_FULLSCREEN_WINDOW) {
	
				al_toggle_display_flag(
					_ts5_data.display[_ts5_status.active_display],
					ALLEGRO_FULLSCREEN_WINDOW, 0);
	
				_ts5_status.display[_ts5_status.active_display].display_mode
					= TS5_WINDOWED;
			}
	
				al_show_native_message_box(
					_ts5_data.display[_ts5_status.active_display],
					"Tscope5 error", "Tscope5 error", outbuf, NULL,
					ALLEGRO_MESSAGEBOX_ERROR);
		}
		else {
			al_show_native_message_box(NULL, "Tscope5 error", "Tscope5 error",
										outbuf, NULL, ALLEGRO_MESSAGEBOX_ERROR);
		}
	
	#endif

	exit(1);
}


////////////////////////////////////////////////////////////////////////////////
//@}
////////////////////////////////////////////////////////////////////////////////
