////////////////////////////////////////////////////////////////////////////////
//     
//    __                                ______
//   / /_______________  ____  ___     / ____/
//  / __/ ___/ ___/ __ \/ __ \/ _ \   /___ )  
// / /_(__  ) /__/ /_/ / /_/ /  __/  ____/ /  
// \__/____/\___/\____/ .___/\___/  /_____/   
//                   /_/                      
//     
/// \file system_internal.c
/// Definitions of internal system functions.
////////////////////////////////////////////////////////////////////////////////

#include "../include/tscope5/system.h"
#include "../include/tscope5/system_internal.h"

/// Is Tscope5 installed?
int _ts5_is_tscope5_installed = 0;

/// Is text log installed?
int _ts5_is_textlog_installed = 0;

/// Settings for the next display that will be opened.
TS5_DISPLAY_STATUS ts5_nextdisplay = 
	{ 0, 640, 360, 0, 0, 60, 0.0, TS5_WINDOWED, TS5_VSYNC_ON};


////////////////////////////////////////////////////////////////////////////////
/// Install Tscope5.
///
/// \param calling_function Name the function that calls 
/// for this check or installation.
///
/// Gathers information about the display adapter(s) of the
/// computer and sets the default parameters for displays, graphics and audio.
///
/// This function is called automatically if necessary.
///
/// If Tscope5 can't be installed the program is aborted.
///
/// This function does not open a display, 
/// that is done by ts5_install_display().
////////////////////////////////////////////////////////////////////////////////
void ts5_install_tscope5(char *calling_function)
{
	int i;

	if (!_ts5_is_tscope5_installed) {

		#ifdef TS5_WINDOWS
			setvbuf(stdout, 0, _IONBF, 0);
					
		#endif

		// system
		_ts5_is_tscope5_installed = 1;
		_ts5_status.loglevel = TS5_LOGLEVEL_3;
		_ts5_data.textlog = NULL;


		// randomizer
		_ts5_data.randomizer = NULL;


		// allegro
		if (!al_install_system(ALLEGRO_VERSION_INT, NULL)) {
			ts5_fatal("%s: could not install Tscope5\n", calling_function);
		}

		ts5_log(TS5_LOGLEVEL_1, "%s: Installing Tscope5\n", calling_function);
		al_inhibit_screensaver(1);
		al_set_separate_blender(ALLEGRO_ADD, ALLEGRO_ALPHA, 
			ALLEGRO_INVERSE_ALPHA, ALLEGRO_ADD, ALLEGRO_ONE, ALLEGRO_ONE);
			
		al_add_new_bitmap_flag(ALLEGRO_NO_PRESERVE_TEXTURE);	
//		al_set_new_display_option(ALLEGRO_SAMPLE_BUFFERS, 1, ALLEGRO_REQUIRE);
//		al_set_new_display_option(ALLEGRO_SAMPLES, 4, ALLEGRO_SUGGEST);


		// display adapters
		_ts5_status.num_display_adapters = al_get_num_video_adapters();

		_ts5_status.display_adapter =
			al_malloc(_ts5_status.num_display_adapters 
			* sizeof(TS5_DISPLAY_ADAPTER_STATUS));

		for (i = 0; i < _ts5_status.num_display_adapters; i++) {

			ALLEGRO_MONITOR_INFO monitorinfo;
			al_get_monitor_info(i, &monitorinfo);
			_ts5_status.display_adapter[i].x1 = monitorinfo.x1;
			_ts5_status.display_adapter[i].x2 = monitorinfo.x2;
			_ts5_status.display_adapter[i].y1 = monitorinfo.y1;
			_ts5_status.display_adapter[i].y2 = monitorinfo.y2;
			_ts5_status.display_adapter[i].w = monitorinfo.x2 - monitorinfo.x1;
			_ts5_status.display_adapter[i].h = monitorinfo.y2 - monitorinfo.y1;
		}

		ts5_nextdisplay.x = -1;
		ts5_nextdisplay.y = -1;


		// displays
		_ts5_status.num_displays = 0;
		_ts5_status.active_display = -1;
		_ts5_status.display = NULL;
		_ts5_data.display = NULL;
		_ts5_data.target = NULL;
		

		// graphics parameters
		_ts5_status.graphics.coordinate_system = TS5_CARTESIAN_COORDINATES;
		_ts5_status.graphics.coordinate_scale = TS5_ABSOLUTE_COORDINATES;

		_ts5_status.graphics.foreground_color = 
			al_map_rgba_f(1.0, 1.0, 1.0, 1.0);
		
		_ts5_status.graphics.background_color = 
			al_map_rgba_f(0.0, 0.0, 0.0, 1.0);
		
		_ts5_status.graphics.drawing_thickness = 0.0;
		_ts5_status.graphics.fill_mode = TS5_FILL_OFF;

		_ts5_status.graphics.font_type = TS5_COURIER;
		_ts5_status.graphics.font_style = TS5_BOLD;
		_ts5_status.graphics.font_size = 20;

		_ts5_status.graphics.font_index =
			_ts5_status.graphics.font_type * TS5_FONTSTYLES * TS5_FONTSIZES +
			_ts5_status.graphics.font_style * TS5_FONTSIZES +
			_ts5_status.graphics.font_size - 8;

		_ts5_status.graphics.text_alignment = TS5_ALIGN_CENTER;

		for (i=0; i<TS5_NFONTS; i++) {
			_ts5_status.graphics.fonts[i] = NULL;
		}
		_ts5_status.graphics.font = NULL;


		// audio parameters
		_ts5_status.audio.samplerate = TS5_44100;
		_ts5_status.audio.channels = TS5_STEREO;
		_ts5_status.audio.depth = TS5_FLOAT;
		_ts5_status.audio.gain = 1.0;

		_ts5_data.audio.voice = NULL;
		_ts5_data.audio.mixer = NULL;


		// timer
		_ts5_status.timer.priority = TS5_NORMAL_PRIORITY;
		_ts5_status.timer.num_defined_buttons = 0;
		_ts5_status.timer.num_active_buttons = 0;

		_ts5_status.timer.voicekey_is_response_device = 0;

		_ts5_status.timer.mouse_is_response_device = 0;

		_ts5_status.timer.keyboard_is_response_device = 0;

		_ts5_status.timer.num_joysticks = 0;
		_ts5_status.timer.joystick_is_response_device = 0;
		_ts5_status.timer.joystick = NULL;

		_ts5_status.timer.num_cedrusboxes = 0;
		_ts5_status.timer.cedrusbox_is_response_device = 0;
		_ts5_status.timer.cedrusbox_is_trigger_device = 0;
		_ts5_status.timer.cedrusbox = NULL;
		_ts5_data.timer.cedrusbox_thread = NULL;

		_ts5_status.timer.num_parports = 3;
		_ts5_status.timer.parport_base_address[0] = TS5_PARPORT_1_BASE;
		_ts5_status.timer.parport_base_address[1] = TS5_PARPORT_2_BASE;
		_ts5_status.timer.parport_base_address[2] = TS5_PARPORT_3_BASE;

		for (i=0; i<_ts5_status.timer.num_parports; i++) {

			_ts5_status.timer.parport_data_register[i] = 
				_ts5_status.timer.parport_base_address[i];

			_ts5_status.timer.parport_status_register[i] = 
				_ts5_status.timer.parport_base_address[i]+1;

			_ts5_status.timer.parport_control_register[i] = 
				_ts5_status.timer.parport_base_address[i]+2;
		}
		
		_ts5_status.timer.parport_is_response_device = 0;
		_ts5_status.timer.parport_is_trigger_device = 0;
		_ts5_status.timer.parport = NULL;
		_ts5_data.timer.parport_thread = NULL;

		_ts5_status.timer.num_serialports = 0;
		_ts5_status.timer.serialport_is_trigger_device = 0;
		_ts5_status.timer.serialport = NULL;
		_ts5_data.timer.serialport_thread = NULL;

		_ts5_data.timer.response_queue = NULL;
		_ts5_data.timer.trigger_queue = NULL;
		_ts5_data.timer.trigger_log = NULL;


		atexit(ts5_uninstall_tscope5);
	}
}


////////////////////////////////////////////////////////////////////////////////
/// Uninstall Tscope5.
///
/// Closes Tscope5, all subsystems and the log window.
///
/// This function is called automatically at the end of the program.
////////////////////////////////////////////////////////////////////////////////
void ts5_uninstall_tscope5()
{
	if (_ts5_is_tscope5_installed) {

		ts5_log(TS5_LOGLEVEL_1, "Uninstalling Tscope5\n");
		al_inhibit_screensaver(0);

		if (_ts5_is_textlog_installed) {
			fflush(_ts5_data.textlog);
			fclose(_ts5_data.textlog);
			_ts5_is_textlog_installed=0;
		}

		al_free(_ts5_status.display_adapter);

		//al_uninstall_system();

		_ts5_is_tscope5_installed = 0;
	}
}

